# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from dataclasses import replace

from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.services.camera import (
    CameraAction,
    CameraEvent,
    CameraStartViewfinderAction,
    CameraStartViewfinderEvent,
    CameraState,
    CameraStopViewfinderAction,
    CameraStopViewfinderEvent,
)
from ubo_app.store.services.keypad import Key, KeypadAction

Action = InitAction | CameraAction


def reducer(
    state: CameraState | None,
    action: Action,
) -> ReducerResult[CameraState, Action, CameraEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return CameraState(
                is_viewfinder_active=False,
            )
        raise InitializationActionError(action)

    if isinstance(action, CameraStartViewfinderAction):
        return CompleteReducerResult(
            state=replace(state, is_viewfinder_active=True),
            events=[
                CameraStartViewfinderEvent(
                    barcode_pattern=action.barcode_pattern,
                ),
            ],
        )

    if isinstance(action, CameraStopViewfinderAction):
        return CompleteReducerResult(
            state=replace(state, is_viewfinder_active=False),
            events=[
                CameraStopViewfinderEvent(),
            ],
        )

    if isinstance(action, KeypadAction):  # noqa: SIM102
        if action.key == Key.BACK:
            return CompleteReducerResult(
                state=replace(state),
                actions=[
                    CameraStopViewfinderAction(),
                ],
            )

    return state



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

import asyncio
import threading
from typing import TYPE_CHECKING, Callable, Coroutine, TypeVarTuple, Unpack

from redux import FinishEvent
from typing_extensions import TypeVar

from ubo_app.constants import DEBUG_MODE
from ubo_app.store import subscribe_event

if TYPE_CHECKING:
    from asyncio import Future, Handle


T = TypeVar('T', infer_variance=True)
Ts = TypeVarTuple('Ts')


class WorkerThread(threading.Thread):
    def __init__(self: WorkerThread) -> None:
        super().__init__()
        self.loop = asyncio.new_event_loop()
        if DEBUG_MODE:
            self.loop.set_debug(enabled=True)

    def run(self: WorkerThread) -> None:
        asyncio.set_event_loop(self.loop)

        subscribe_event(FinishEvent, self.stop)
        self.loop.run_forever()

    def run_task(self: WorkerThread, task: Coroutine) -> Handle:
        return self.loop.call_soon_threadsafe(self.loop.create_task, task)

    def run_in_executor(
        self: WorkerThread,
        executor: object,
        task: Callable[[Unpack[Ts]], T],
        *args: Unpack[Ts],
    ) -> Future[T]:
        return self.loop.run_in_executor(executor, task, *args)

    def stop(self: WorkerThread) -> None:
        self.loop.call_soon_threadsafe(self.loop.stop)


thread = WorkerThread()
thread.start()


def _create_task(task: Coroutine) -> Handle:
    return thread.run_task(task)


def _run_in_executor(
    executer: object,
    task: Callable[[Unpack[Ts]], T],
    *args: Unpack[Ts],
) -> Future[T]:
    return thread.run_in_executor(executer, task, *args)


_ = _create_task, _run_in_executor



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from redux import BaseAction, BaseEvent, Immutable


class CameraAction(BaseAction):
    ...


class CameraStartViewfinderAction(CameraAction):
    barcode_pattern: str | None


class CameraStopViewfinderAction(CameraAction):
    ...


class CameraBarcodeAction(CameraAction):
    code: str
    match: dict[str, str | None]


class CameraEvent(BaseEvent):
    ...


class CameraStartViewfinderEvent(CameraEvent):
    barcode_pattern: str | None


class CameraStopViewfinderEvent(CameraEvent):
    ...


class CameraState(Immutable):
    is_viewfinder_active: bool



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from typing import TYPE_CHECKING, Literal

from kivy.core.window import Keyboard, Window, WindowBase

from ubo_app.store.services.keypad import (
    Key,
    KeypadKeyPressAction,
)
from ubo_app.store.services.sound import (
    SoundDevice,
    SoundToggleMuteStatusAction,
)

if TYPE_CHECKING:
    Modifier = Literal['ctrl', 'alt', 'meta', 'shift']


def on_keyboard(
    _: WindowBase,
    key: int,
    _scancode: int,
    _codepoint: str,
    modifier: list[Modifier],
) -> None:
    """Handle keyboard events."""
    from ubo_app.store import dispatch

    if modifier == []:
        if key == Keyboard.keycodes['up']:
            dispatch(KeypadKeyPressAction(key=Key.UP))
        elif key == Keyboard.keycodes['down']:
            dispatch(KeypadKeyPressAction(key=Key.DOWN))
        elif key == Keyboard.keycodes['1']:
            dispatch(KeypadKeyPressAction(key=Key.L1))
        elif key == Keyboard.keycodes['2']:
            dispatch(KeypadKeyPressAction(key=Key.L2))
        elif key == Keyboard.keycodes['3']:
            dispatch(KeypadKeyPressAction(key=Key.L3))
        elif key == Keyboard.keycodes['left']:
            dispatch(KeypadKeyPressAction(key=Key.BACK))
        elif key == Keyboard.keycodes['m']:
            from ubo_app.store import dispatch

            dispatch(
                SoundToggleMuteStatusAction(
                    device=SoundDevice.INPUT,
                ),
            )


def init_service() -> None:
    Window.bind(on_keyboard=on_keyboard)



"""Setup the service."""
from __future__ import annotations

from datetime import datetime, timezone

import adafruit_pct2075
import adafruit_veml7700
import board
from kivy.clock import Clock

from ubo_app.store import dispatch
from ubo_app.store.services.sensors import Sensor, SensorsReportReadingAction


def read_sensors(_: float | None = None) -> None:
    """Read the sensor."""
    i2c = board.I2C()
    temperature_sensor = adafruit_pct2075.PCT2075(i2c, address=0x48)
    temperature = temperature_sensor.temperature
    light_sensor = adafruit_veml7700.VEML7700(i2c, address=0x10)
    light = light_sensor.lux
    dispatch(
        SensorsReportReadingAction(
            sensor=Sensor.TEMPERATURE,
            reading=temperature,
            timestamp=datetime.now(tz=timezone.utc),
        ),
        SensorsReportReadingAction(
            sensor=Sensor.LIGHT,
            reading=light,
            timestamp=datetime.now(tz=timezone.utc),
        ),
    )


def init_service() -> None:
    """Initialize the service."""
    Clock.schedule_interval(read_sensors, 1)
    read_sensors()



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from typing import Sequence

from redux import BaseAction, Immutable


class IconState(Immutable):
    symbol: str
    color: str
    priority: int
    id: str | None


class StatusIconsState(Immutable):
    icons: Sequence[IconState]


class StatusIconsAction(BaseAction):
    ...


class StatusIconsRegisterAction(StatusIconsAction):
    icon: str
    color: str = 'white'
    priority: int = 0
    id: str | None = None



# ruff: noqa: D100, D101, D102, D103, D104, D107
from pathlib import Path

from headless_kivy_pi.constants import BYTES_PER_PIXEL

from ubo_app.utils.fake import Fake

IS_RPI = Path('/etc/rpi-issue').exists()
if not IS_RPI:
    import sys

    sys.modules['RPi'] = Fake()

from RPi import GPIO  # pyright: ignore [reportMissingImports]  # noqa: E402


def initialize_board() -> None:
    if not IS_RPI:
        return

    GPIO.setmode(GPIO.BCM)
    GPIO.setup(17, GPIO.IN, pull_up_down=GPIO.PUD_DOWN)


def turn_off_screen() -> None:
    if not IS_RPI:
        return

    GPIO.setup(26, GPIO.OUT)
    GPIO.output(26, GPIO.LOW)

    from headless_kivy_pi import HeadlessWidget

    display = HeadlessWidget._display  # noqa: SLF001
    data = [0] * HeadlessWidget.width * HeadlessWidget.height * BYTES_PER_PIXEL
    display._block(0, 0, HeadlessWidget.width - 1, HeadlessWidget.height - 1, data)  # noqa: SLF001


def turn_on_screen() -> None:
    if not IS_RPI:
        return

    GPIO.setmode(GPIO.BCM)
    GPIO.setup(26, GPIO.OUT)
    GPIO.output(26, GPIO.HIGH)



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import datetime
from functools import cached_property
from typing import TYPE_CHECKING, Any, Sequence

from kivy.clock import Clock
from kivy.metrics import dp
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.label import Label
from kivy.uix.stencilview import StencilView
from kivy.uix.widget import Widget
from ubo_gui.app import UboApp

from ubo_app.store import autorun

if TYPE_CHECKING:
    from ubo_app.store.status_icons import IconState


class MenuAppFooter(UboApp):
    @cached_property
    def temperature_widget(self: MenuAppFooter) -> BoxLayout:
        layout = BoxLayout(
            orientation='horizontal',
            spacing=0,
            padding=0,
            size_hint=(None, 1),
        )

        temperature = Label(font_size=dp(14), size_hint=(None, 1), valign='middle')
        temperature.bind(
            texture_size=lambda temperature, texture_size: setattr(
                temperature,
                'width',
                texture_size[0],
            )
            or setattr(layout, 'width', temperature.width + dp(12)),
        )

        @autorun(lambda state: state.sensors.temperature.value)
        def set_value(value: float | None = None) -> None:
            if value is None:
                temperature.text = '-'
            else:
                temperature.text = f'{value:0.1f}°C'

        icon = Label(
            text='device_thermostat',
            color='#ffffff',
            font_name='material_symbols',
            font_size=dp(16),
            font_features='fill=0',
            size_hint=(None, 1),
            width=dp(12),
        )
        layout.add_widget(icon)
        layout.add_widget(temperature)

        return layout

    @cached_property
    def light_widget(self: MenuAppFooter) -> Label:
        light = Label(
            text='light_mode',
            color='#ffffff',
            font_name='material_symbols',
            font_size=dp(16),
            font_features='fill=0',
            size_hint=(None, 1),
            width=dp(16),
        )
        light.bind(
            texture_size=lambda light, texture_size: setattr(
                light,
                'width',
                texture_size[0],
            ),
        )

        @autorun(lambda state: state.sensors.light.value)
        def set_value(value: float | None = None) -> None:
            if value is None:
                light.color = (0.5, 0, 0, 1)
            else:
                v = min(value, 140) / 140
                light.color = (1, 1, 1, v)

        return light

    @cached_property
    def clock_widget(self: MenuAppFooter) -> Label:
        clock = Label(font_size=dp(20), size_hint=(None, 1), valign='middle')
        clock.bind(
            texture_size=lambda clock, texture_size: setattr(
                clock,
                'width',
                texture_size[0],
            ),
        )
        local_timzone = datetime.datetime.now(datetime.timezone.utc).astimezone().tzinfo

        def now() -> datetime.datetime:
            return datetime.datetime.now(local_timzone)

        def set_value(_: int | None = None) -> None:
            clock.text = now().strftime('%H:%M')

        def update(_: int | None = None) -> None:
            set_value()
            now_ = now()
            Clock.schedule_once(
                update,
                60 - now_.second - now_.microsecond / 1000000 + 0.05,
            )

        update()

        return clock

    @cached_property
    def footer(self: MenuAppFooter) -> Widget | None:
        layout = BoxLayout()

        normal_footer_layout = BoxLayout(orientation='horizontal', spacing=0, padding=0)
        normal_footer_layout.add_widget(
            Label(
                text='reply',
                font_name='material_symbols',
                font_size=dp(20),
                font_features='fill=1',
                size_hint=(None, 1),
            ),
        )
        normal_footer_layout.add_widget(Widget(size_hint=(1, 1)))

        home_footer_layout = BoxLayout(orientation='horizontal', spacing=0, padding=0)

        home_footer_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))
        home_footer_layout.add_widget(self.clock_widget)
        home_footer_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))
        home_footer_layout.add_widget(self.temperature_widget)
        home_footer_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))
        home_footer_layout.add_widget(self.light_widget)
        home_footer_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))

        icons_widget = StencilView(size_hint=(1, 1))
        home_footer_layout.add_widget(icons_widget)

        icons_layout = BoxLayout(
            orientation='horizontal',
            spacing=dp(4),
            padding=0,
        )
        icons_widget.add_widget(icons_layout)

        def set_icons_layout_x(*_: list[Any]) -> None:
            icons_layout.x = icons_widget.x + icons_widget.width - icons_layout.width

        icons_widget.bind(
            width=set_icons_layout_x,
            x=set_icons_layout_x,
            height=icons_layout.setter('height'),
            y=icons_layout.setter('y'),
        )
        icons_layout.bind(
            width=set_icons_layout_x,
            x=set_icons_layout_x,
        )

        @autorun(lambda state: state.status_icons.icons)
        def render_icons(selector_result: Sequence[IconState]) -> None:
            icons = selector_result
            icons_layout.clear_widgets()
            for icon in list(reversed(icons))[:4]:
                label = Label(
                    text=icon.symbol,
                    color=icon.color,
                    font_name='material_symbols',
                    font_size=dp(20),
                    font_features='fill=0',
                    size_hint=(None, 1),
                    width=dp(22),
                )
                icons_layout.add_widget(label)
            icons_layout.add_widget(Widget(size_hint=(None, 1), width=dp(2)))
            icons_layout.bind(minimum_width=icons_layout.setter('width'))

        @autorun(lambda state: state.main.path)
        def handle_depth_change(path: Sequence[str]) -> None:
            is_fullscreen = (
                'TODO' in path
            )  # TODO(sassanh): Check if the application is fullscreen
            if not is_fullscreen:
                if normal_footer_layout in layout.children:
                    layout.remove_widget(normal_footer_layout)
                    layout.add_widget(home_footer_layout)
            elif home_footer_layout in layout.children:
                layout.remove_widget(home_footer_layout)
                layout.add_widget(normal_footer_layout)

        layout.add_widget(home_footer_layout)

        return layout



# ruff: noqa: D100, D101, D102, D103, D104, D105, D107, N999
from __future__ import annotations

from typing import Any, Sequence

from constants import get_signal_icon
from debouncer import DebounceOptions, debounce
from kivy.properties import BooleanProperty
from ubo_gui.menu.types import (
    ActionItem,
    ApplicationItem,
    HeadlessMenu,
    Item,
    SubMenuItem,
)
from ubo_gui.prompt import PromptWidget
from wifi_manager import (
    connect_wireless_connection,
    disconnect_wireless_connection,
    forget_wireless_connection,
    get_active_connection_ssid,
    get_wifi_device,
)

from ubo_app.store import autorun, dispatch
from ubo_app.store.services.wifi import (
    ConnectionState,
    WiFiState,
    WiFiUpdateRequestAction,
)
from ubo_app.utils.async_ import create_task

from .create_wireless_connection import CreateWirelessConnectionPage


class WiFiConnectionPage(PromptWidget):
    ssid: str
    is_active = BooleanProperty(defaultvalue=None)

    def first_option_callback(self: WiFiConnectionPage) -> None:
        if self.is_active:
            create_task(disconnect_wireless_connection())
        else:
            create_task(connect_wireless_connection(self.ssid))
        dispatch(
            WiFiUpdateRequestAction(
                reset=True,
            ),
        )

    def second_option_callback(self: WiFiConnectionPage) -> None:
        create_task(forget_wireless_connection(self.ssid))
        self.dispatch('on_close')
        dispatch(
            WiFiUpdateRequestAction(reset=True),
        )

    def update(self: WiFiConnectionPage, *_: tuple[Any, ...]) -> None:
        self.first_option_background_color = (
            PromptWidget.first_option_background_color.defaultvalue
        )
        if self.is_active:
            self.first_option_label = 'Disconnect'
            self.first_option_icon = 'link_off'
            self.icon = 'wifi'
        else:
            self.first_option_label = 'Connect'
            self.first_option_icon = 'link'
            self.icon = 'wifi_off'

    def __init__(self: WiFiConnectionPage, **kwargs: object) -> None:
        super().__init__(**kwargs, items=None)
        self.prompt = f'SSID: {self.ssid}'
        self.icon = 'hourglass_top'
        self.first_option_background_color = 'black'
        self.first_option_label = ''
        self.first_option_is_short = False
        self.second_option_label = 'Delete'
        self.second_option_icon = 'delete'
        self.second_option_is_short = False

        self.bind(is_active=self.update)

        @debounce(
            wait=0.5,
            options=DebounceOptions(leading=False, trailing=True, time_window=2),
        )
        async def update_status() -> None:
            self.is_active = await get_active_connection_ssid() == self.ssid

        create_task(update_status())

        async def listener() -> None:
            wifi_device = await get_wifi_device()
            if not wifi_device:
                return

            async for _ in wifi_device.properties_changed:
                create_task(update_status())

        create_task(listener())


@autorun(lambda state: state.wifi)
def wireless_connection_items(wifi_state: WiFiState) -> Sequence[Item]:
    if not wifi_state:
        return []

    def wifi_network_creator(ssid: str) -> type[WiFiConnectionPage]:
        class WiFiNetworkPageWithSSID(WiFiConnectionPage):
            def __init__(self: WiFiNetworkPageWithSSID, **kwargs: object) -> None:
                self.ssid = ssid
                self.title = None
                super().__init__(**kwargs)

        return WiFiNetworkPageWithSSID

    icons = {
        ConnectionState.CONNECTED: 'link',
        ConnectionState.DISCONNECTED: 'link-off',
        ConnectionState.CONNECTING: 'pending',
        ConnectionState.UNKNOWN: 'question',
    }
    return (
        [
            ApplicationItem(
                label=connection.ssid,
                application=wifi_network_creator(connection.ssid),
                icon=get_signal_icon(connection.signal_strength)
                if connection.state == ConnectionState.DISCONNECTED
                else icons[connection.state],
            )
            for connection in wifi_state.connections
        ]
        if wifi_state.connections is not None
        else [ActionItem(label='Loading...', action=lambda: None)]
    )


def list_connections() -> HeadlessMenu:
    dispatch(WiFiUpdateRequestAction())
    return HeadlessMenu(
        title='Wi-Fi Connections',
        items=wireless_connection_items,
    )


WiFiMainMenu = SubMenuItem(
    label='WiFi',
    icon='wifi',
    sub_menu=HeadlessMenu(
        title='WiFi Settings',
        items=[
            ApplicationItem(
                label='Add',
                icon='wifi_add',
                application=CreateWirelessConnectionPage,
            ),
            ActionItem(
                label='Select',
                icon='list',
                action=list_connections,
            ),
        ],
    ),
)



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from enum import Enum, auto

from immutable import Immutable
from redux import BaseAction, BaseEvent

UPDATE_MANAGER_NOTIFICATION_ID = 'ubo:update_manager'


class UpdateManagerAction(BaseAction):
    ...


class UpdateManagerSetVersionsAction(UpdateManagerAction):
    flash_notification: bool
    latest_version: str
    current_version: str


class UpdateManagerSetStatusAction(UpdateManagerAction):
    status: UpdateStatus


class UpdateManagerEvent(BaseEvent):
    ...


class UpdateManagerCheckEvent(UpdateManagerEvent):
    ...


class UpdateManagerUpdateEvent(UpdateManagerEvent):
    ...


class UpdateStatus(Enum):
    """Update status enum."""

    CHECKING = auto()
    FAILED_TO_CHECK = auto()
    UP_TO_DATE = auto()
    OUTDATED = auto()
    UPDATING = auto()


class UpdateManagerState(Immutable):
    """Version store."""

    current_version: str | None = None
    latest_version: str | None = None
    update_status: UpdateStatus = UpdateStatus.CHECKING



"""Update manager module."""
from __future__ import annotations

import asyncio
import importlib.metadata
from pathlib import Path

import aiohttp
from kivy import shutil
from ubo_gui.constants import DANGER_COLOR, SUCCESS_COLOR
from ubo_gui.menu.types import ActionItem, Item

from ubo_app.constants import INSTALLATION_PATH
from ubo_app.logging import logger
from ubo_app.store import autorun, dispatch
from ubo_app.store.services.notifications import (
    Chime,
    Notification,
    NotificationDisplayType,
    NotificationsAddAction,
)
from ubo_app.store.update_manager import (
    UpdateManagerSetStatusAction,
    UpdateManagerSetVersionsAction,
    UpdateManagerState,
    UpdateStatus,
)
from ubo_app.store.update_manager.reducer import ABOUT_MENU_PATH

CURRENT_VERSION = importlib.metadata.version('ubo_app')


async def check_version() -> None:
    """Check for updates."""
    logger.info('Checking for updates...')

    # Check PyPI server for the latest version
    import requests

    try:
        async with aiohttp.ClientSession() as session, session.get(
            'https://pypi.org/pypi/ubo-app/json',
            timeout=5,
        ) as response:
            if response.status != requests.codes.ok:
                logger.error('Failed to check for updates')
                return
            data = await response.json()
            latest_version = data['info']['version']

            dispatch(
                with_state=lambda state: UpdateManagerSetVersionsAction(
                    flash_notification=state is None
                    or state.main.path[:3] != ABOUT_MENU_PATH,
                    current_version=CURRENT_VERSION,
                    latest_version=latest_version,
                ),
            )
    except Exception as exception:  # noqa: BLE001
        logger.error('Failed to check for updates', exc_info=exception)
        dispatch(UpdateManagerSetStatusAction(status=UpdateStatus.FAILED_TO_CHECK))
        return


async def update() -> None:
    """Update the Ubo app."""
    logger.info('Updating Ubo app...')

    async def download_files() -> None:
        target_path = Path(f'{INSTALLATION_PATH}/_update/')
        shutil.rmtree(target_path, ignore_errors=True)
        target_path.mkdir(parents=True, exist_ok=True)

        process = await asyncio.create_subprocess_exec(
            '/usr/bin/env',
            'pip',
            'download',
            '--dest',
            target_path,
            'ubo-app[default]',
            'setuptools',
            'wheel',
        )
        await process.wait()
        if process.returncode != 0:
            msg = 'Failed to download packages'
            raise RuntimeError(msg)

        target_path.joinpath('update_is_ready').touch()

    try:
        await download_files()

        await asyncio.sleep(2)

        process = await asyncio.create_subprocess_exec(
            '/usr/bin/env',
            'systemctl',
            'reboot',
            '-i',
        )
        await process.wait()
    except Exception as exception:  # noqa: BLE001
        logger.error('Failed to update', exc_info=exception)
        dispatch(
            NotificationsAddAction(
                notification=Notification(
                    title='Failed to update',
                    content='Failed to update',
                    display_type=NotificationDisplayType.FLASH,
                    color=DANGER_COLOR,
                    icon='security_update_warning',
                    chime=Chime.FAILURE,
                ),
            ),
            UpdateManagerSetStatusAction(status=UpdateStatus.CHECKING),
        )
        return


@autorun(lambda state: state.update_manager)
def about_menu_items(state: UpdateManagerState) -> list[Item]:
    """Get the update menu items."""
    if state.update_status is UpdateStatus.CHECKING:
        return [
            ActionItem(
                label='Checking for updates...',
                action=lambda: None,
                icon='update',
                background_color='#00000000',
            ),
        ]
    if state.update_status is UpdateStatus.FAILED_TO_CHECK:
        return [
            ActionItem(
                label='Failed to check for updates',
                action=lambda: dispatch(
                    UpdateManagerSetStatusAction(status=UpdateStatus.CHECKING),
                ),
                icon='security_update_warning',
                background_color=DANGER_COLOR,
            ),
        ]
    if state.update_status is UpdateStatus.UP_TO_DATE:
        return [
            ActionItem(
                label='Already up to date!',
                action=lambda: None,
                icon='security_update_good',
                background_color=SUCCESS_COLOR,
                color='#000000',
            ),
        ]
    if state.update_status is UpdateStatus.OUTDATED:
        return [
            ActionItem(
                label=f'Update to v{state.latest_version}',
                action=lambda: dispatch(
                    UpdateManagerSetStatusAction(status=UpdateStatus.UPDATING),
                ),
                icon='system_update',
            ),
        ]
    if state.update_status is UpdateStatus.UPDATING:
        return [
            ActionItem(
                label='Updating...',
                action=lambda: None,
                icon='update',
                background_color='#00000000',
            ),
        ]
    return []



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import os
from distutils.util import strtobool
from pathlib import Path

USERNAME = 'ubo'
INSTALLATION_PATH = '/opt/ubo'
DEBUG_MODE = strtobool(os.environ.get('UBO_DEBUG', 'False')) == 1
LOG_LEVEL = os.environ.get('UBO_LOG_LEVEL', 'DEBUG' if DEBUG_MODE else None)
GUI_LOG_LEVEL = os.environ.get('UBO_GUI_LOG_LEVEL', 'DEBUG' if DEBUG_MODE else None)
SERVICES_PATH = (
    os.environ.get('UBO_SERVICES_PATH', '').split(':')
    if os.environ.get('UBO_SERVICES_PATH')
    else []
)
SOCKET_PATH = Path('/run/ubo').joinpath('system_manager.sock').as_posix()
DOCKER_INSTALLATION_LOCK_FILE = Path('/var/run/ubo/docker_installation.lock')



# ruff: noqa: D100, D101, D102, D103, D104, D105, D107
from __future__ import annotations

from types import ModuleType
from typing import Any, Generator, Iterator

from ubo_app.logging import logger


class Fake(ModuleType):
    def __init__(self: Fake) -> None:
        super().__init__('')

    def __getattr__(self: Fake, attr: str) -> Fake | str:
        logger.verbose(
            'Accessing fake attribute of a `Fake` insta',
            extra={'attr': attr},
        )
        if attr == '__file__':
            return ''
        return self

    def __getitem__(self: Fake, key: str) -> Fake:
        logger.verbose(
            'Accessing fake item of a `Fake` instance',
            extra={'key': key},
        )
        return self

    def __call__(self: Fake, *args: object, **kwargs: dict[str, Any]) -> Fake:
        logger.verbose(
            'Calling a `Fake` instance',
            extra={'args_': args, 'kwargs': kwargs},
        )
        return self

    def __await__(self: Fake) -> Generator[Fake | None, Any, Any]:
        yield None
        return self

    def __iter__(self: Fake) -> Iterator[Fake]:
        return iter([self])

    def __enter__(self: Fake) -> Fake:  # noqa: PYI034
        return self

    def __exit__(self: Fake, *_: object) -> None:
        pass



"""Docker store types."""
from __future__ import annotations

from dataclasses import field
from enum import Enum, auto

from redux import BaseAction, BaseCombineReducerState, BaseEvent, Immutable


class DockerStatus(Enum):
    """Docker status."""

    UNKNOWN = auto()
    NOT_INSTALLED = auto()
    INSTALLING = auto()
    NOT_RUNNING = auto()
    RUNNING = auto()
    ERROR = auto()


class ImageStatus(Enum):
    """Image status."""

    NOT_AVAILABLE = auto()
    FETCHING = auto()
    AVAILABLE = auto()
    CREATED = auto()
    RUNNING = auto()
    ERROR = auto()


class DockerAction(BaseAction):
    """Docker action."""


class DockerSetStatusAction(DockerAction):
    """Docker set status action."""

    status: DockerStatus


class DockerImageAction(DockerAction):
    """Docker image action."""

    image: str


class DockerImageSetStatusAction(DockerImageAction):
    """Docker image action."""

    status: ImageStatus
    ports: list[str] | None = None


class DockerEvent(BaseEvent):
    """Docker event."""


class DockerServiceState(Immutable):
    """Docker service state."""

    status: DockerStatus = DockerStatus.UNKNOWN


class DockerImageEvent(DockerEvent):
    """Docker image event."""

    image: str


class ImageState(Immutable):
    """Image state."""

    id: str
    label: str
    icon: str
    path: str
    status: ImageStatus = ImageStatus.NOT_AVAILABLE
    ports: list[str] = field(default_factory=list)

    @property
    def is_fetching(self: ImageState) -> bool:
        """Check if image is available."""
        return self.status == ImageStatus.FETCHING

    @property
    def is_available(self: ImageState) -> bool:
        """Check if image is available."""
        return self.status in [ImageStatus.AVAILABLE, ImageStatus.RUNNING]

    @property
    def is_running(self: ImageState) -> bool:
        """Check if image is running."""
        return self.status == ImageStatus.RUNNING


class DockerState(BaseCombineReducerState):
    """Docker state."""

    service: DockerServiceState

    def __getattribute__(self: DockerState, name: str) -> ImageState:
        """Set type for random attributes of DockerState."""
        return super().__getattribute__(name)



"""Setup the service."""
from __future__ import annotations

import asyncio
import contextlib
import socket
import subprocess
from dataclasses import fields
from pathlib import Path
from typing import Callable

import docker
import docker.errors
from docker.models.containers import Container
from docker.models.images import Image
from reducer import IMAGE_IDS, IMAGES
from ubo_gui.menu.types import ActionItem, HeadedMenu, HeadlessMenu, Item, SubMenuItem

from ubo_app.constants import DOCKER_INSTALLATION_LOCK_FILE, SOCKET_PATH
from ubo_app.logging import logger
from ubo_app.store import autorun, dispatch
from ubo_app.store.main import RegisterSettingAppAction
from ubo_app.store.services.docker import (
    DockerImageSetStatusAction,
    DockerSetStatusAction,
    DockerState,
    DockerStatus,
    ImageState,
    ImageStatus,
)
from ubo_app.utils.async_ import create_task, run_in_executor


def send_command(command: bytes) -> None:
    """Send a command to the system manager socket."""
    server_socket = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM)
    try:
        server_socket.connect(SOCKET_PATH)
    except Exception as exception:  # noqa: BLE001
        logger.error('Unable to connect to the socket', exc_info=exception)
        return
    else:
        server_socket.send(command)
    finally:
        server_socket.close()


def install_docker() -> None:
    """Install Docker."""
    dispatch(DockerSetStatusAction(status=DockerStatus.INSTALLING))

    if Path(SOCKET_PATH).exists():
        send_command(b'docker install')


def run_docker() -> None:
    """Install Docker."""
    send_command(b'docker start')

    dispatch(DockerSetStatusAction(status=DockerStatus.UNKNOWN))


def stop_docker() -> None:
    """Install Docker."""
    subprocess.run(
        ['/usr/bin/env', 'systemctl', 'stop', 'docker'],  # noqa: S603
        check=False,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )

    dispatch(DockerSetStatusAction(status=DockerStatus.UNKNOWN))


def find_container(client: docker.DockerClient, *, image: str) -> Container | None:
    """Find a container."""
    for container in client.containers.list(all=True):
        if not isinstance(container, Container):
            continue

        with contextlib.suppress(docker.errors.DockerException):
            container_image = container.image
            if isinstance(container_image, Image) and image in container_image.tags:
                return container
    return None


@autorun(lambda state: state.docker.service.status)
def setup_menu(status: DockerStatus) -> HeadedMenu:
    """Get the menu items for the Docker service."""
    if status in [
        DockerStatus.UNKNOWN,
        DockerStatus.INSTALLING,
        DockerStatus.NOT_INSTALLED,
        DockerStatus.NOT_RUNNING,
    ]:

        async def check_docker() -> None:
            """Check if Docker is installed."""
            process = await asyncio.create_subprocess_exec(
                '/usr/bin/env',
                'which',
                'docker',
            )
            await process.wait()
            is_installed = process.returncode == 0

            is_running = False
            with contextlib.suppress(docker.errors.DockerException):
                docker_client = docker.from_env()
                is_running = docker_client.ping()
                docker_client.close()

            if is_running:
                dispatch(DockerSetStatusAction(status=DockerStatus.RUNNING))
            elif is_installed:
                dispatch(DockerSetStatusAction(status=DockerStatus.NOT_RUNNING))
            elif Path(DOCKER_INSTALLATION_LOCK_FILE).exists():
                dispatch(DockerSetStatusAction(status=DockerStatus.INSTALLING))
            else:
                dispatch(DockerSetStatusAction(status=DockerStatus.NOT_INSTALLED))

        create_task(check_docker())

    title = 'Setup Docker'
    if status == DockerStatus.UNKNOWN:
        return HeadedMenu(
            title=title,
            heading='Checking',
            sub_heading='Checking Docker service status',
            items=[],
        )
    if status == DockerStatus.NOT_INSTALLED:
        return HeadedMenu(
            title=title,
            heading='Docker is not Installed',
            sub_heading='Install it to enjoy the power of Docker on your Ubo pod',
            items=[
                ActionItem(
                    label='Install Docker',
                    icon='system_update',
                    action=install_docker,
                ),
            ],
        )
    if status == DockerStatus.INSTALLING:
        return HeadedMenu(
            title=title,
            heading='Installing...',
            sub_heading='Docker is being installed',
            items=[],
        )
    if status == DockerStatus.NOT_RUNNING:
        return HeadedMenu(
            title=title,
            heading='Docker is not Running',
            sub_heading='Run it to enjoy the power of Docker on your Ubo pod',
            items=[
                ActionItem(
                    label='Start Docker',
                    icon='play_arrow',
                    action=run_docker,
                ),
            ],
        )
    if status == DockerStatus.RUNNING:
        return HeadedMenu(
            title=title,
            heading='Docker is Running',
            sub_heading='Enjoy the power of Docker on your Ubo pod',
            items=[
                ActionItem(
                    label='Stop Docker',
                    icon='stop',
                    action=stop_docker,
                ),
            ],
        )
    if status == DockerStatus.ERROR:
        return HeadedMenu(
            title=title,
            heading='Docker Error',
            sub_heading='Please check the logs for more information',
            items=[],
        )

    msg = f'Unknown status: {status}'
    raise ValueError(msg)


def image_menu_generator(image_id: str) -> Callable[[], Callable[[], HeadedMenu]]:
    """Get the menu items for the Docker service."""

    def check_image(path: str) -> None:
        def act() -> None:
            logger.debug('Checking image', extra={'image': image_id, 'path': path})
            docker_client = docker.from_env()
            try:
                docker_client.images.get(path)
                logger.debug('Image found', extra={'image': image_id, 'path': path})

                container = find_container(docker_client, image=path)
                if container:
                    if container.status == 'running':
                        logger.debug(
                            'Container running image found',
                            extra={'image': image_id, 'path': path},
                        )
                        dispatch(
                            DockerImageSetStatusAction(
                                image=image_id,
                                status=ImageStatus.RUNNING,
                                ports=[
                                    f'{i["HostIp"]}:{i["HostPort"]}'
                                    for i in container.ports.values()
                                    for i in i
                                ],
                            ),
                        )
                        return
                    logger.debug(
                        "Container for the image found, but it's not running",
                        extra={'image': image_id, 'path': path},
                    )
                    dispatch(
                        DockerImageSetStatusAction(
                            image=image_id,
                            status=ImageStatus.CREATED,
                        ),
                    )
                    return

                logger.debug(
                    'Container running image not found',
                    extra={'image': image_id, 'path': path},
                )
                dispatch(
                    DockerImageSetStatusAction(
                        image=image_id,
                        status=ImageStatus.AVAILABLE,
                    ),
                )
            except docker.errors.ImageNotFound as exception:
                logger.debug(
                    'Image not found',
                    extra={'image': image_id, 'path': path},
                    exc_info=exception,
                )
                dispatch(
                    DockerImageSetStatusAction(
                        image=image_id,
                        status=ImageStatus.NOT_AVAILABLE,
                    ),
                )
            except docker.errors.DockerException as exception:
                logger.debug(
                    'Image error',
                    extra={'image': image_id, 'path': path},
                    exc_info=exception,
                )
                dispatch(
                    DockerImageSetStatusAction(
                        image=image_id,
                        status=ImageStatus.ERROR,
                    ),
                )
            finally:
                docker_client.close()

        run_in_executor(None, act)

    @autorun(lambda state: getattr(state.docker, image_id))
    def image_items(image: ImageState) -> list[Item]:
        def fetch_image() -> None:
            def act() -> None:
                dispatch(
                    DockerImageSetStatusAction(
                        image=image_id,
                        status=ImageStatus.FETCHING,
                    ),
                )
                try:
                    logger.debug('Fetching image', extra={'image': image.path})
                    docker_client = docker.from_env()
                    response = docker_client.api.pull(
                        image.path,
                        stream=True,
                        decode=True,
                    )
                    for line in response:
                        logger.verbose(
                            'Image pull',
                            extra={'image': image.path, 'line': line},
                        )
                    logger.debug('Image fetched', extra={'image': image.path})
                    check_image(image.path)
                    docker_client.close()
                except docker.errors.DockerException as exception:
                    logger.debug(
                        'Image error',
                        extra={'image': image.path},
                        exc_info=exception,
                    )
                    dispatch(
                        DockerImageSetStatusAction(
                            image=image_id,
                            status=ImageStatus.ERROR,
                        ),
                    )

            run_in_executor(None, act)

        def remove_image() -> None:
            def act() -> None:
                docker_client = docker.from_env()
                docker_client.images.remove(image.path, force=True)
                check_image(image.path)
                docker_client.close()

            run_in_executor(None, act)

        def run_container() -> None:
            def act() -> None:
                docker_client = docker.from_env()
                container = find_container(docker_client, image=image.path)
                if container:
                    if container.status != 'running':
                        container.start()
                else:
                    docker_client.containers.run(
                        image.path,
                        publish_all_ports=True,
                        detach=True,
                        volumes=IMAGES[image_id].volumes,
                        **(
                            {'ports': IMAGES[image_id].ports}
                            if IMAGES[image_id].ports
                            else {}
                        ),
                    )
                check_image(image.path)
                docker_client.close()

            run_in_executor(None, act)

        def stop_container() -> None:
            def act() -> None:
                docker_client = docker.from_env()
                container = find_container(docker_client, image=image.path)
                if container and container.status != 'exited':
                    container.stop()
                check_image(image.path)
                docker_client.close()

            run_in_executor(None, act)

        def remove_container() -> None:
            def act() -> None:
                docker_client = docker.from_env()
                container = find_container(docker_client, image=image.path)
                if container:
                    container.remove(v=True, force=True)
                check_image(image.path)
                docker_client.close()

            run_in_executor(None, act)

        items: list[Item] = []

        if image.status == ImageStatus.NOT_AVAILABLE:
            items.append(
                ActionItem(
                    label='Fetch',
                    icon='download',
                    action=fetch_image,
                ),
            )
        elif image.status == ImageStatus.FETCHING:
            items.append(
                ActionItem(
                    label='Stop',
                    icon='stop',
                    action=remove_image,
                ),
            )
        elif image.status == ImageStatus.AVAILABLE:
            items.extend(
                [
                    ActionItem(
                        label='Start',
                        icon='play_arrow',
                        action=run_container,
                    ),
                    ActionItem(
                        label='Remove image',
                        icon='delete',
                        action=remove_image,
                    ),
                ],
            )
        elif image.status == ImageStatus.CREATED:
            items.extend(
                [
                    ActionItem(
                        label='Start',
                        icon='play_arrow',
                        action=run_container,
                    ),
                    ActionItem(
                        label='Remove container',
                        icon='delete',
                        action=remove_container,
                    ),
                ],
            )
        elif image.status == ImageStatus.RUNNING:
            items.append(
                ActionItem(
                    label='Stop',
                    icon='stop',
                    action=stop_container,
                ),
            )
            items.append(
                SubMenuItem(
                    label='ports',
                    icon='category',
                    sub_menu=HeadlessMenu(
                        title='Ports',
                        items=[Item(label=i, icon='category') for i in image.ports],
                    ),
                ),
            )
        elif image.status == ImageStatus.ERROR:
            items.append(
                ActionItem(
                    label='Check status',
                    icon='refresh',
                    action=lambda: check_image(image.path),
                ),
            )

        return items

    @autorun(lambda state: getattr(state.docker, image_id))
    def image_menu(
        image: ImageState,
    ) -> HeadedMenu:
        """Get the menu items for the Docker service."""
        return HeadedMenu(
            title=f'Docker - {image.label}',
            heading=image.label,
            sub_heading={
                ImageStatus.NOT_AVAILABLE: 'Image needs to be fetched',
                ImageStatus.FETCHING: 'Image is being fetched',
                ImageStatus.AVAILABLE: 'Image is ready but container is not running',
                ImageStatus.CREATED: 'Container is created but not running',
                ImageStatus.RUNNING: 'Container is running',
                ImageStatus.ERROR: 'Image has an error, please check the logs',
            }[image.status],
            items=image_items,
        )

    def image_action() -> Callable[[], HeadedMenu]:
        """Get the menu items for the Docker service."""
        check_image(IMAGES[image_id].path)

        return image_menu

    return image_action


image_menus = {image_id: image_menu_generator(image_id) for image_id in IMAGE_IDS}


def setup_menu_action() -> Callable[[], HeadedMenu]:
    """Get the menu items for the Docker service."""
    dispatch(DockerSetStatusAction(status=DockerStatus.UNKNOWN))
    return setup_menu


@autorun(lambda state: state.docker)
def docker_menu_items(state: DockerState) -> list[Item]:
    """Get the menu items for the Docker service."""
    items: list[Item] = [
        ActionItem(
            label='Setup Docker',
            icon='manufacturing',
            action=setup_menu_action,
        ),
    ]

    if state.service.status == DockerStatus.RUNNING:
        items.append(
            SubMenuItem(
                label='Docker Containers',
                icon='category',
                sub_menu=HeadlessMenu(
                    title='Docker Containers',
                    items=[
                        ActionItem(
                            label=getattr(state, image_id).label,
                            icon=getattr(state, image_id).icon,
                            action=image_menus[image_id],
                        )
                        for image_id in (field.name for field in fields(state))
                        if image_id not in (field.name for field in fields(DockerState))
                    ],
                ),
            ),
        )

    return items


def docker_menu_item_action() -> HeadlessMenu:
    """Get the menu items for the Docker service."""
    return HeadlessMenu(
        title='Docker',
        items=docker_menu_items,
    )


docker_main_menu = ActionItem(
    label='Docker',
    icon='D',
    action=docker_menu_item_action,
)


def init_service() -> None:
    """Initialize the service."""
    dispatch(RegisterSettingAppAction(menu_item=docker_main_menu))



# pyright: reportMissingImports=false
# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

import logging
import math
import time
from enum import StrEnum
from typing import TYPE_CHECKING, Literal

import board
from immutable import Immutable

from ubo_app.store.services.keypad import Key, KeypadKeyPressAction
from ubo_app.store.services.sound import SoundDevice, SoundSetMuteStatusAction

INT_EXPANDER = 5  # GPIO PIN index that receives interrupt from AW9523

ButtonStatus = Literal['pressed', 'released']


class KeypadError(Exception):
    ...


class ButtonName(StrEnum):
    TOP_LEFT = 'top_left'
    MIDDLE_LEFT = 'middle_left'
    BOTTOM_LEFT = 'bottom_left'
    UP = 'up'
    DOWN = 'down'
    BACK = 'back'
    HOME = 'home'
    MIC = 'mic'


class ButtonEvent(Immutable):
    status: ButtonStatus
    timestamp: float


def init_service() -> None:
    import adafruit_aw9523
    from RPi import GPIO

    if TYPE_CHECKING:
        from adafruit_bus_device import i2c_device
        from adafruit_register.i2c_struct import UnaryStruct

    class Event(Immutable):
        inputs: UnaryStruct
        timestamp: float

    class KeypadStatus:
        """Class to keep track of button status.

        0: "TOP_LEFT": Top left button
        1: "MIDDLE_LEFT": Middle left button
        2: "BOTTOM_LEFT": Bottom left button
        3: "BACK": Button with back arrow label (under LCD)
        4: "HOME": Button with home label (Under LCD)
        5: "UP": Right top button with upward arrow label
        6: "DOWN": Right bottom button with downward arrow label
        7: "MIC": Microphone mute switch
        """

        buttons: dict[ButtonName, ButtonEvent]

        def __init__(self: KeypadStatus) -> None:
            self.buttons = {
                button_name: ButtonEvent(status='released', timestamp=time.time())
                for button_name in ButtonName
            }

        def update_status(
            self: KeypadStatus,
            button_name: ButtonName,
            new_status: Literal['pressed', 'released'],
        ) -> None:
            if new_status not in {'pressed', 'released'}:
                msg = 'Invalid status'
                raise KeypadError(msg)
            if button_name in self.buttons:
                self.buttons[button_name] = ButtonEvent(
                    status=new_status,
                    timestamp=time.time(),
                )

        def get_status(
            self: KeypadStatus,
            button_name: ButtonName,
        ) -> Literal['pressed', 'released']:
            if button_name in self.buttons:
                return self.buttons[button_name].status
            msg = 'Invalid button name'
            raise KeypadError(msg)

        def get_timestamp(self: KeypadStatus, button_name: ButtonName) -> float:
            if button_name in self.buttons:
                return self.buttons[button_name].timestamp
            msg = 'Invalid button name'
            raise KeypadError(msg)

        def get_label(self: KeypadStatus, index: int) -> ButtonName:
            if index > len(ButtonName):
                msg = 'Invalid index'
                raise KeypadError(msg)
            return list(ButtonName)[index]

    class Keypad:
        """Class to handle keypad events."""

        event_queue: list[Event]
        aw: adafruit_aw9523.AW9523 | None
        inputs: UnaryStruct | None

        def __init__(self: Keypad) -> None:
            """Initialize a Keypad.

            Initializes various parameters including
            loggers and button names.
            """
            self.logger = logging.getLogger('keypad')
            self.logger.setLevel(logging.WARNING)
            self.logger.debug('Initialising keypad...')
            self.event_queue = []
            self.aw = None
            self.inputs = None
            self.bus_address = 0x58
            self.model = 'aw9523'
            self.enabled = True
            self.buttons = KeypadStatus()
            self.index = 0
            self.init_i2c()

        def clear_interrupt_flags(self: Keypad, i2c: i2c_device.I2CDevice) -> None:
            # Write to both registers to reset the interrupt flag
            buffer = bytearray(2)
            buffer[0] = 0x00
            buffer[1] = 0x00
            i2c.write(buffer)
            i2c.write_then_readinto(buffer, buffer, out_end=1, in_start=1)

            time.sleep(0.1)
            buffer[0] = 0x01
            buffer[1] = 0x00
            i2c.write(buffer)
            i2c.write_then_readinto(buffer, buffer, out_end=1, in_start=1)
            time.sleep(0.1)

        def disable_interrupt_for_higher_bits(
            self: Keypad,
            i2c: i2c_device.I2CDevice,
        ) -> None:
            # disable interrupt for higher bits
            buffer = bytearray(2)
            buffer[0] = 0x06
            buffer[1] = 0x00
            i2c.write(buffer)
            i2c.write_then_readinto(buffer, buffer, out_end=1, in_start=1)

            buffer[0] = 0x07
            buffer[1] = 0xFF
            i2c.write(buffer)
            i2c.write_then_readinto(buffer, buffer, out_end=1, in_start=1)

        def init_i2c(self: Keypad) -> None:
            # connect to the I2C bus
            GPIO.setmode(GPIO.BCM)
            i2c = board.I2C()
            # Set this to the GPIO of the interrupt:
            GPIO.setup(INT_EXPANDER, GPIO.IN, pull_up_down=GPIO.PUD_UP)
            try:
                # Search for the GPIO expander address on the I2C bus
                self.aw = adafruit_aw9523.AW9523(i2c, self.bus_address)
                new_i2c = self.aw.i2c_device
            except Exception:
                self.bus_address = False
                self.logger.exception('Failed to initialize I2C Bus on address 0x58')
                raise

            # Perform soft reset of the expander
            self.aw.reset()
            # Set first 8 low significant bits (register 1) to input
            self.aw.directions = 0xFF00
            time.sleep(1)

            # The code below, accessing the GPIO expander registers directly via i2c
            # was created as a workaround of the reset the interrupt flag.
            self.clear_interrupt_flags(new_i2c)
            self.disable_interrupt_for_higher_bits(new_i2c)
            # reset interrupts again
            self.clear_interrupt_flags(new_i2c)

            # read register values
            self.inputs = self.aw.inputs
            self.logger.debug(
                'Initializing inputs',
                extra={'inputs': f'{self.inputs:016b}'},
            )
            self.event_queue = [Event(inputs=self.inputs, timestamp=time.time())]
            time.sleep(0.5)

            # Enable interrupt on the GPIO expander
            GPIO.add_event_detect(
                INT_EXPANDER,
                GPIO.FALLING,
                callback=self.key_press_cb,
                bouncetime=1,
            )

        def key_press_cb(self: Keypad, _: object) -> None:
            """Handle key press dispatched by GPIO interrupt.

                This is callback function that gets triggers
             if any change is detected on keypad buttons
             states.

             In this callback, we look at the state
             change to see which button was pressed or leased.

            Parameters
            ----------
            _channel: int
                GPIO channel that triggered the callback
                NOT USED currently
            """
            if self.aw is None:
                return
            # read register values
            self.inputs = self.aw.inputs
            event = Event(inputs=self.inputs, timestamp=time.time())
            # append the event to the queue. The queue has a depth of 2 and
            # keeps the current and last event.
            self.event_queue.append(event)
            self.logger.info(self.event_queue)
            self.logger.debug('Current Inputs', extra={'inputs': f'{self.inputs:016b}'})
            previos_event = self.event_queue.pop(0)
            self.logger.debug(
                'Previous Inputs',
                extra={'inputs': f'{previos_event.inputs:016b}'},
            )
            # XOR the last recorded input values with the current input values
            # to see which bits have changed. Technically there can only be one
            # bit change in every callback
            change_mask = previos_event.inputs ^ self.inputs
            self.logger.debug('Change', extra={'change_mask': f'{change_mask:016b}'})
            # use the change mask to see if the button was the change was
            # falling (1->0) indicating a pressed action
            # or risign (0->1) indicating a release action
            self.index = (int)(math.log2(change_mask))
            self.logger.info('button index', extra={'button_index': self.index})
            # Check for rising edge or falling edge action (press or release)
            self.button_label = self.buttons.get_label(self.index)
            if (previos_event.inputs & change_mask) == 0:
                self.logger.info(
                    'Button pressed',
                    extra={'button': str(self.index), 'label': self.button_label},
                )

                # calculate how long the button was held down
                # and print the time
                last_time_stamp = self.buttons.get_timestamp(self.button_label)
                held_down_time = time.time() - last_time_stamp
                self.logger.info(
                    'Button was pressed down',
                    extra={'held_down_time': held_down_time},
                )

                self.buttons.update_status(self.button_label, 'released')
                self.on_button_event(self.button_label, 'released')

            else:
                self.logger.info(
                    'Button released',
                    extra={'button': str(self.index), 'label': self.button_label},
                )
                self.buttons.update_status(self.button_label, 'pressed')
                self.on_button_event(self.button_label, 'pressed')

            self.logger.info(self.buttons.buttons)

        def on_button_event(
            self: Keypad,
            button_pressed: ButtonName,
            button_status: ButtonStatus,
        ) -> None:
            from ubo_app.store import dispatch

            if button_status == 'pressed':
                if button_pressed == ButtonName.UP:
                    dispatch(
                        KeypadKeyPressAction(
                            key=Key.UP,
                        ),
                    )
                elif button_pressed == ButtonName.DOWN:
                    dispatch(
                        KeypadKeyPressAction(
                            key=Key.DOWN,
                        ),
                    )
                elif button_pressed == ButtonName.TOP_LEFT:
                    dispatch(
                        KeypadKeyPressAction(
                            key=Key.L1,
                        ),
                    )
                elif button_pressed == ButtonName.MIDDLE_LEFT:
                    dispatch(
                        KeypadKeyPressAction(
                            key=Key.L2,
                        ),
                    )
                elif button_pressed == ButtonName.BOTTOM_LEFT:
                    dispatch(
                        KeypadKeyPressAction(
                            key=Key.L3,
                        ),
                    )
                elif button_pressed == ButtonName.BACK:
                    dispatch(
                        KeypadKeyPressAction(
                            key=Key.BACK,
                        ),
                    )
                elif button_pressed == ButtonName.HOME:
                    dispatch(
                        KeypadKeyPressAction(
                            key=Key.HOME,
                        ),
                    )
            if button_pressed == ButtonName.MIC:
                from ubo_app.store import dispatch

                dispatch(
                    SoundSetMuteStatusAction(
                        device=SoundDevice.INPUT,
                        mute=button_status == 'pressed',
                    ),
                )

    Keypad()



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from constants import ETHERNET_STATE_ICON_ID, ETHERNET_STATE_ICON_PRIORITY
from debouncer import DebounceOptions, debounce
from ethernet_manager import get_ethernet_device, get_ethernet_device_state

from ubo_app.store import dispatch
from ubo_app.store.services.ethernet import GlobalEthernetState
from ubo_app.store.status_icons import StatusIconsRegisterAction
from ubo_app.utils.async_ import create_task


@debounce(
    wait=0.5,
    options=DebounceOptions(leading=True, trailing=False, time_window=2),
)
async def update_ethernet_icon() -> None:
    state = await get_ethernet_device_state()
    dispatch(
        StatusIconsRegisterAction(
            icon={
                GlobalEthernetState.CONNECTED: 'link',
                GlobalEthernetState.DISCONNECTED: 'link_off',
                GlobalEthernetState.PENDING: 'settings_ethernet',
                GlobalEthernetState.NEEDS_ATTENTION: ('settings_ethernet'),
                GlobalEthernetState.UNKNOWN: 'indeterminate_question_box',
            }[state],
            priority=ETHERNET_STATE_ICON_PRIORITY,
            id=ETHERNET_STATE_ICON_ID,
        ),
    )


async def setup_listeners() -> None:
    ethernet_device = await get_ethernet_device()
    if not ethernet_device:
        return

    async for _ in ethernet_device.properties_changed:
        create_task(update_ethernet_icon())


def init_service() -> None:
    create_task(update_ethernet_icon())
    create_task(setup_listeners())



# ruff: noqa: D100, D101, D102, D103, D104, D107, N999
from __future__ import annotations

from dataclasses import replace

from constants import SOUND_MIC_STATE_ICON_ID, SOUND_MIC_STATE_ICON_PRIORITY
from redux import (
    CompleteReducerResult,
    InitAction,
    InitializationActionError,
    ReducerResult,
)

from ubo_app.store.services.notifications import Chime
from ubo_app.store.services.sound import (
    SoundAction,
    SoundChangeVolumeAction,
    SoundDevice,
    SoundPlayChimeAction,
    SoundPlayChimeEvent,
    SoundSetMuteStatusAction,
    SoundSetVolumeAction,
    SoundState,
    SoundToggleMuteStatusAction,
)
from ubo_app.store.status_icons import StatusIconsRegisterAction

Action = InitAction | SoundAction | StatusIconsRegisterAction


def reducer(  # noqa: C901, PLR0912
    state: SoundState | None,
    action: Action,
) -> ReducerResult[SoundState, Action, SoundPlayChimeEvent]:
    if state is None:
        if isinstance(action, InitAction):
            return SoundState(
                playback_volume=0.5,
                is_playback_mute=False,
                capture_volume=0.5,
                is_capture_mute=False,
            )
        raise InitializationActionError(action)

    if isinstance(action, SoundSetVolumeAction):
        if action.device == SoundDevice.OUTPUT:
            return CompleteReducerResult(
                state=replace(state, playback_volume=action.volume),
                events=[
                    SoundPlayChimeEvent(name=Chime.VOLUME_CHANGE),
                ],
            )
        if action.device == SoundDevice.INPUT:
            return replace(state, capture_volume=action.volume)
    elif isinstance(action, SoundChangeVolumeAction):
        if action.device == SoundDevice.OUTPUT:
            return CompleteReducerResult(
                state=state,
                actions=[
                    SoundSetVolumeAction(
                        device=SoundDevice.OUTPUT,
                        volume=min(
                            max(state.playback_volume + action.amount, 0),
                            1,
                        ),
                    ),
                ],
            )
        if action.device == SoundDevice.INPUT:
            return replace(
                state,
                capture_volume=min(
                    max(state.capture_volume + action.amount, 0),
                    1,
                ),
            )
    elif isinstance(action, SoundSetMuteStatusAction):
        if action.device == SoundDevice.OUTPUT:
            return replace(state, is_playback_mute=action.mute)
        if action.device == SoundDevice.INPUT:
            return CompleteReducerResult(
                state=replace(state, is_capture_mute=action.mute),
                actions=[
                    StatusIconsRegisterAction(
                        icon='mic_off' if action.mute else 'mic',
                        priority=SOUND_MIC_STATE_ICON_PRIORITY,
                        id=SOUND_MIC_STATE_ICON_ID,
                    ),
                ],
            )
    elif isinstance(action, SoundToggleMuteStatusAction):
        return CompleteReducerResult(
            state=state,
            actions=[
                SoundSetMuteStatusAction(
                    mute=not state.is_playback_mute
                    if action.device == SoundDevice.OUTPUT
                    else not state.is_capture_mute,
                    device=action.device,
                ),
            ],
        )
    elif isinstance(action, SoundPlayChimeAction):
        return CompleteReducerResult(
            state=state,
            events=[
                SoundPlayChimeEvent(name=action.name),
            ],
        )
    return state



"""Implement `setup_service` function to set up and enable systemd service."""
from __future__ import annotations

import grp
import os
import pwd
import subprocess
from pathlib import Path
from typing import Literal, TypedDict

from ubo_app.constants import INSTALLATION_PATH, USERNAME
from ubo_app.logging import logger


class Service(TypedDict):
    """System service metadata."""

    name: str
    template: str
    scope: Literal['system', 'user']


services: list[Service] = [
    {
        'name': 'ubo-system',
        'template': 'system',
        'scope': 'system',
    },
    {
        'name': 'ubo-update',
        'template': 'update',
        'scope': 'system',
    },
    {
        'name': 'ubo-app',
        'template': 'app',
        'scope': 'user',
    },
]

uid = pwd.getpwnam(USERNAME).pw_uid
gid = grp.getgrnam(USERNAME).gr_gid


def bootstrap() -> None:
    """Create the service files and enable the services."""
    for service in services:
        if service['scope'] == 'user':
            path = Path(f'/home/{USERNAME}/.config/systemd/user')
            path.mkdir(parents=True, exist_ok=True)
            service_file_path = (
                f'/home/{USERNAME}/.config/systemd/user/{service["name"]}.service'
            )
            while path != Path(f'/home/{USERNAME}'):
                os.chown(path, uid, gid)
                path = path.parent
        elif service['scope'] == 'system':
            service_file_path = f'/etc/systemd/system/{service["name"]}.service'
        else:
            logger.error(
                f"Service '{service['name']}' has an invalid scope: {service['scope']}",
            )
            return

        # Ensure we have the required permissions
        if os.geteuid() != 0:
            logger.error('This script needs to be run with root privileges.')
            return

        template = (
            Path(__file__)
            .parent.joinpath(f'services/{service["template"]}.service')
            .open()
            .read()
        )

        content = template.replace(
            '{{INSTALLATION_PATH}}',
            INSTALLATION_PATH,
        ).replace(
            '{{USERNAME}}',
            USERNAME,
        )

        # Write the service content to the file
        with Path(service_file_path).open('w') as file:
            file.write(content)
            if service['scope'] == 'user':
                os.chown(service_file_path, uid, gid)

    subprocess.run(['/usr/bin/env', 'loginctl', 'enable-linger', USERNAME], check=True)  # noqa: S603
    subprocess.run(
        [  # noqa: S603
            '/usr/bin/env',
            'sudo',
            f'XDG_RUNTIME_DIR=/run/user/{uid}',
            '-u',
            USERNAME,
            'systemctl',
            '--user',
            'daemon-reload',
        ],
        check=True,
    )
    subprocess.run(
        ['/usr/bin/env', 'systemctl', 'daemon-reload'],  # noqa: S603
        check=True,
    )

    for service in services:
        # Enable the service to start on boot
        if service['scope'] == 'user':
            subprocess.run(
                [  # noqa: S603
                    '/usr/bin/env',
                    'sudo',
                    f'XDG_RUNTIME_DIR=/run/user/{uid}',
                    '-u',
                    USERNAME,
                    'systemctl',
                    '--user',
                    'enable',
                    service['name'],
                ],
                check=True,
            )
        elif service['scope'] == 'system':
            subprocess.run(
                ['/usr/bin/env', 'systemctl', 'enable', service['name']],  # noqa: S603
                check=True,
            )

        logger.info(
            f"Service '{service['name']}' has been created and enabled.",
        )

    with Path('/etc/polkit-1/rules.d/50-ubo.rules').open('w') as file:
        file.write(
            Path(__file__)
            .parent.joinpath('polkit-reboot.rules')
            .open()
            .read()
            .replace('{{INSTALLATION_PATH}}', INSTALLATION_PATH)
            .replace('{{USERNAME}}', USERNAME),
        )

    subprocess.run(
        [Path(__file__).parent.joinpath('install_wm8960.sh').as_posix()],  # noqa: S603
        check=True,
    )



# ruff: noqa: D100, D101, D102, D103, D104, D107
from __future__ import annotations

from enum import StrEnum
from typing import Sequence

from redux import BaseAction, BaseEvent, Immutable


class WiFiType(StrEnum):
    WEP = 'WEP'
    WPA = 'WPA'
    WPA2 = 'WPA2'
    nopass = 'NOPASS'


class ConnectionState(StrEnum):
    CONNECTED = 'Connected'
    CONNECTING = 'Connecting'
    DISCONNECTED = 'Disconnected'
    UNKNOWN = 'Unknown'


class GlobalWiFiState(StrEnum):
    CONNECTED = 'Connected'
    DISCONNECTED = 'Disconnected'
    PENDING = 'Pending'
    NEEDS_ATTENTION = 'Needs Attention'
    UNKNOWN = 'Unknown'


class WiFiConnection(Immutable):
    ssid: str
    state: ConnectionState = ConnectionState.UNKNOWN
    signal_strength: int = 0
    password: str | None = None
    type: WiFiType | None = None
    hidden: bool = False


class WiFiAction(BaseAction):
    ...


class WiFiUpdateAction(WiFiAction):
    connections: Sequence[WiFiConnection]
    state: GlobalWiFiState
    current_connection: WiFiConnection | None


class WiFiUpdateRequestAction(WiFiAction):
    reset: bool = False


class WiFiEvent(BaseEvent):
    ...


class WiFiCreateEvent(WiFiEvent):
    connection: WiFiConnection


class WiFiUpdateRequestEvent(WiFiEvent):
    ...


class WiFiState(Immutable):
    connections: Sequence[WiFiConnection] | None
    state: GlobalWiFiState
    current_connection: WiFiConnection | None
